'''Show basic usage of Robust coil combination for bSSFP.'''

from time import time

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import multivariate_normal
from phantominator import shepp_logan

from ssfp import bssfp, robustcc, gs_recon

def _gaussian_csm(sx, sy, ncoil, sigma=1):
    '''Simple coil model for demo.'''
    X, Y = np.meshgrid(
        np.linspace(-1, 1, sx), np.linspace(-1, 1, sy))
    pos = np.stack((X[..., None], Y[..., None]), axis=-1)
    csm = np.empty((sx, sy, ncoil))
    cov = [[sigma, 0], [0, sigma]]
    for ii in range(ncoil):
        mu = [np.cos(ii/ncoil*np.pi*2), np.sin(ii/ncoil*2*np.pi)]
        csm[..., ii] = multivariate_normal(mu, cov).pdf(pos)
    return csm + 1j*csm

if __name__ == '__main__':

    # Coil combine params
    include_full_robustcc = False # this might take a while to run

    # Sim params
    N, nc, npcs = 256, 8, 4
    pcs = np.linspace(0, 2*np.pi, npcs, endpoint=False)
    M0, T1, T2 = shepp_logan((N, N, 1), MR=True, zlims=(-.25, .25))
    M0, T1, T2 = np.squeeze(M0), np.squeeze(T1), np.squeeze(T2)

    # Linear off resonance
    TR, alpha = 6e-3, np.deg2rad(30)
    df, _ = np.meshgrid(
        np.linspace(-1/TR, 1/TR, N),
        np.linspace(-1/TR, 1/TR, N))

    # Generate coil images
    csm = _gaussian_csm(N, N, nc)
    data = np.empty((N, N, npcs, nc), dtype=np.complex64)
    for ii in range(nc):
        res = bssfp(
            T1, T2, TR=TR, alpha=alpha, field_map=df, phase_cyc=pcs,
            M0=M0, delta_cs=0, phi_rf=np.angle(csm[None, ..., ii]),
            phi_edd=0, phi_drift=0, target_pc_axis=-1)
        data[..., ii] = np.abs(csm[..., None, ii])*res

    # Do coil-by-coil recon
    res_cbc = np.empty((N, N, nc), dtype=np.complex64)
    t0 = time()
    for ii in range(nc):
        res_cbc[..., ii] = gs_recon(data[..., ii], pc_axis=-1)
    res_cbc = np.sqrt(np.sum(np.abs(res_cbc)**2, axis=-1))
    print('Took %g seconds to do coil-by-coil recon' % (time() - t0))

    # Do robust coil combine then recon
    t0 = time()
    res_rcc_simple = robustcc(data, method='simple')
    res_rcc_simple = np.abs(gs_recon(res_rcc_simple, pc_axis=-1))
    print('Took %g seconds to do simple robustcc recon' % (time()-t0))

    if include_full_robustcc:
        t0 = time()
        res_rcc_full = robustcc(data, method='full', mask=M0 > 0)
        res_rcc_full = np.abs(gs_recon(res_rcc_full, pc_axis=-1))
        print(
            'Took %g seconds to do full robustcc recon' % (time()-t0))

    # Take a look
    nx, ny = 1, 2
    vmax = np.max(np.concatenate((res_cbc, res_rcc_simple)).flatten())
    if include_full_robustcc:
        vmax = np.maximum(vmax, np.max(res_rcc_full.flatten()))
        ny += 1
    plt_args = {
        'vmin': 0,
        'vmax': vmax,
        'cmap': 'gray',
    }
    plt.subplot(nx, ny, 1)
    plt.imshow(res_cbc, **plt_args)
    plt.title('Coil-by-coil')
    plt.axis('off')

    plt.subplot(nx, ny, 2)
    plt.imshow(res_rcc_simple, **plt_args)
    plt.title('Simple')
    plt.axis('off')

    if include_full_robustcc:
        plt.subplot(nx, ny, 3)
        plt.imshow(res_rcc_full, **plt_args)
        plt.title('Full')
        plt.axis('off')

    plt.show()



'''GRE simulation.'''

import logging

import numpy as np

from .utils import ernst

def spoiled_gre(T1, T2star, TR, TE, alpha=None, M0=1):
    '''Spoiled, steady state GRE contrast simulation.

    Magnitude image simulations.

    Parameters
    ----------
    T1 : array_like
        longitudinal exponential decay time constant (in sec).
    T2star : array_like
        Effective transverse exponential decay time constant (in sec).
    TR : float
        repetition time (in sec).
    TE : float
        echo time (in sec).
    alpha : array_like, optional
        flip angle (in rad).
    M0 : array_like, optional
        proton density.

    Returns
    -------
    S : array_like
        Simulated magnitude spoiled GRE image.

    Notes
    -----
    alpha=None will use the Ernst angle.

    Assuming a longitudinal steady-state and perfect spoiling. Note
    that dependence is on T2* rather than T2 because SE/STE formation
    is suppressed by spoiling and the signal is generated by gradient
    refocusing of an FID.

    Implements equation [14.8] from [1]_.

    References
    ----------
    .. [1] Notes from Bernstein, M. A., King, K. F., & Zhou, X. J.
           (2004). Handbook of MRI pulse sequences. Elsevier.
    '''

    if alpha is None:
        logging.warning(
            'Ernst angle will computed for every T1 value, this may '
            'or may not be what you wanted')
        alpha = ernst(TR, T1)

    # Make sure we don't divide by zero
    idx1 = np.nonzero(T1)
    idx2 = np.nonzero(T2star)
    E1 = np.zeros(T1.shape)
    E1[idx1] = np.exp(-TR/T1[idx1])
    E2 = np.zeros(T2star.shape)
    E2[idx2] = np.exp(-TE/T2star[idx2])

    # Do the thing:
    return M0*np.sin(alpha)*(1 - E1)*E2/(1 - np.cos(alpha)*E1)



'''Extend PLANET recon across harmonics.

Notes
-----
This may be possible due to the fact that spectra/ellipses are very
similar between harmonics (not the same, but might be close enough).

This does not work very well...
'''

import numpy as np
import matplotlib.pyplot as plt
from phantominator import shepp_logan
from tqdm import tqdm

from ssfp import bssfp, planet

if __name__ == '__main__':

    N, npcs = 128, 16
    pcs = np.linspace(0, 2*np.pi, npcs, endpoint=False)
    M0, T1, T2 = shepp_logan((N, N, 1), MR=True, zlims=(-.25, .25))
    M0, T1, T2 = np.squeeze(M0), np.squeeze(T1), np.squeeze(T2)

    # Linear off resonance
    TR0 = 3e-3
    TR1 = 6e-3
    alpha = np.deg2rad(60)
    TRmax = np.maximum(TR0, TR1)
    TRmin = np.minimum(TR0, TR1)
    df, _ = np.meshgrid(
        np.linspace(-1/TRmax, 1/TRmax, N),
        np.linspace(-1/TRmax, 1/TRmax, N))

    # Simulate acquisition
    data = np.empty((N, N, npcs), dtype=np.complex64)

    data0 = bssfp(
        T1, T2, TR=TR0, alpha=alpha, field_map=df,
        phase_cyc=pcs[0::2], M0=M0, delta_cs=0, phi_rf=0, phi_edd=0,
        phi_drift=0)
    data0 = np.moveaxis(data0, 0, -1)
    data[..., 0::2] = data0

    data1 = bssfp(
        T1, T2, TR=TR1, alpha=alpha, field_map=df,
        phase_cyc=pcs[1::2], M0=M0, delta_cs=0, phi_rf=0, phi_edd=0,
        phi_drift=0)
    data1 = np.moveaxis(data1, 0, -1)
    data[..., 1::2] = data1

    Meff = np.zeros(M0.shape, dtype=np.complex64)
    T1est = np.zeros(T1.shape, dtype=np.complex64)
    T2est = np.zeros(T2.shape, dtype=np.complex64)
    mask = M0 > 0
    for idx in tqdm(
            np.argwhere(mask),
            total=np.sum(mask.flatten()),
            leave=False):
        try:
            Meff[tuple(idx)], T1est[tuple(idx)], T2est[tuple(idx)] = planet(
                data[tuple(idx) + (slice(None),)], alpha, TRmin,
                T1[tuple(idx)])
        except AssertionError:
            pass

    plt.imshow(np.abs(T2est))
    plt.show()



    # # Try recon
    # res = gs_recon(data, pc_axis=-1)

    # plt.imshow(np.abs(res))
    # plt.show()



'''Show basic usage of GS solution.'''

from time import perf_counter

import numpy as np
import matplotlib.pyplot as plt
from skimage.metrics import normalized_root_mse
from skimage.restoration import unwrap_phase
from phantominator import shepp_logan
from ssfp import bssfp, planet


if __name__ == '__main__':

    # Shepp-Logan
    N, nslices, npcs = 128, 2, 8  # 2 slices just to show we can
    M0, T1, T2 = shepp_logan((N, N, nslices), MR=True, zlims=(-.25, 0))

    # Simulate bSSFP acquisition with linear off-resonance
    TR, alpha = 3e-3, np.deg2rad(15)
    pcs = np.linspace(0, 2*np.pi, npcs, endpoint=False)
    df, _ = np.meshgrid(
        np.linspace(-1/TR, 1/TR, N),
        np.linspace(-1/TR, 1/TR, N))
    sig = np.empty((npcs,) + T1.shape, dtype='complex')
    for sl in range(nslices):
        sig[..., sl] = bssfp(
            T1[..., sl], T2[..., sl], TR, alpha, field_map=df,
            phase_cyc=pcs, M0=M0[..., sl])

    # Do T1, T2 mapping for each pixel
    mask = np.abs(M0) > 1e-8

    # Make it noisy
    np.random.seed(0)
    sig += 1e-5*(np.random.normal(0, 1, sig.shape) +
                 1j*np.random.normal(0, 1, sig.shape))*mask

    # Do the thing
    t0 = perf_counter()
    Mmap, T1est, T2est, dfest = planet(sig, alpha, TR, mask=mask, pc_axis=0)
    print('Took %g sec to run PLANET' % (perf_counter() - t0))

    # Look at a single slice
    sl = 0
    T1est = T1est[..., sl]
    T2est = T2est[..., sl]
    dfest = dfest[..., sl]
    T1 = T1[..., sl]
    T2 = T2[..., sl]
    mask = mask[..., sl]

    # Simple phase unwrapping of off-resonance estimate
    dfest = unwrap_phase(dfest*2*np.pi*TR)/(2*np.pi*TR)

    nx, ny = 3, 3
    plt.subplot(nx, ny, 1)
    plt.imshow(T1*mask)
    plt.title('T1 Truth')
    plt.axis('off')

    plt.subplot(nx, ny, 2)
    plt.imshow(T1est)
    plt.title('T1 est')
    plt.axis('off')

    plt.subplot(nx, ny, 3)
    plt.imshow(T1*mask - T1est)
    plt.title('NRMSE: %g' % normalized_root_mse(T1, T1est))
    plt.axis('off')

    plt.subplot(nx, ny, 4)
    plt.imshow(T2*mask)
    plt.title('T2 Truth')
    plt.axis('off')

    plt.subplot(nx, ny, 5)
    plt.imshow(T2est)
    plt.title('T2 est')
    plt.axis('off')

    plt.subplot(nx, ny, 6)
    plt.imshow(T2*mask - T2est)
    plt.title('NRMSE: %g' % normalized_root_mse(T2, T2est))
    plt.axis('off')

    plt.subplot(nx, ny, 7)
    plt.imshow(df*mask)
    plt.title('df Truth')
    plt.axis('off')

    plt.subplot(nx, ny, 8)
    plt.imshow(dfest)
    plt.title('df est')
    plt.axis('off')

    plt.subplot(nx, ny, 9)
    plt.imshow(df*mask - dfest)
    plt.title('NRMSE: %g' % normalized_root_mse(df*mask, dfest))
    plt.axis('off')

    plt.show()



'''Python implementation of PLANET algorithm.

Notes
-----
Halir ellipse performs much faster! Somewhat different results though
between fast guaranteed and Halir -- it would be interesting to do an
analysis of the difference between the two methods and truth.
'''

import numpy as np
from ellipsinator import fast_guaranteed_ellipse_estimate as ellipse_fit


def planet(I, alpha, TR, T1_guess=None, pcs=None, mask=None, pc_axis=-1, ret_all=False):
    '''Simultaneous T1, T2 mapping using phase‐cycled bSSFP.

    Parameters
    ----------
    I : array_like
        Complex phase-cycled bSSFP images. The data may be
        arbitrarily dimensional. The phase-cycle axis (pc_axis)
        holds the phase-cycle data.
    alpha : float or array_like
        Flip angle (in rad).
    TR : float
        Repetition time (in sec).
    T1_guess : float, optional
        Estimate of expected T1 value (in sec). If None, 1 sec
        is used as the default.
    pcs : array_like, optional
        The RF phase-cycle increments used.  If `None`,
        evenly spaced phase-cycles are assumed on [0, 2pi).
    mask : array_like or None, optional
        Which pixels of I to reconstruct.  If None, the mask
        reconstructs each pixel that has nonzero PC data.
    pc_axis : int, optional
        The axis that holds the phase-cycle data.
    ret_all : bool, optional
        Return all matrices return by original PLANET.

    Returns
    -------
    Meff : array_like
        Effective magnetization amplitude (arbitrary units).
    T1 : array_like
        Estimate of T1 values (in sec).
    T2 : array_like
        Estimate of T2 values (in sec).
    df : array_like
        Estimate of off-resonance (in Hz).
    phi : array_like, optional
    Xc : array_like, optional
    Yc : array_like, optional
    A : array_like, optional
    B : array_like, optional

    Notes
    -----
    Requires at least 6 phase cycles to fit the ellipse.  The ellipse
    fitting method they use (and which is implemented here) may not
    be the best method, but it is quick.  Could add more options for
    fitting in the future.

    Implements algorithm described in [1]_.

    References
    ----------
    .. [1] Shcherbakova, Yulia, et al. "PLANET: an ellipse fitting
           approach for simultaneous T1 and T2 mapping using
           phase‐cycled balanced steady‐state free precession."
           Magnetic resonance in medicine 79.2 (2018): 711-722.
    '''
    # Phase cycles to the back
    I = np.moveaxis(I, pc_axis, -1)
    npcs = I.shape[-1]
    sh = I.shape[:-1]

    if pcs is None:
        pcs = np.linspace(0, 2*np.pi, npcs, endpoint=False)
    else:
        assert len(pcs) == npcs, "pcs and data must match!"

    # alpha can either be a scalar or array
    # Choose an intial estimate for T1
    if T1_guess is None:
        T1_guess = 1  # 1sec arbitrariliy

    # Fit all ellipses that are nonzero
    if mask is None:
        recon_idx = np.nonzero(
            np.sum(np.abs(I).reshape((-1, npcs)), axis=-1))[0]
    else:
        recon_idx = np.nonzero(mask.flatten())[0]

    I = np.take(I.reshape((-1, npcs)), recon_idx, axis=0)
    ellipse_coefs = ellipse_fit(I)

    # Filter out all the fitting failures
    failure_idx = np.where(np.sum(np.abs(ellipse_coefs), axis=-1) == 0)[0]
    recon_idx = np.setdiff1d(recon_idx, failure_idx)
    ellipse_coefs = np.delete(ellipse_coefs, failure_idx, axis=0)

    # Find ellipse rotation
    A, B, C, D, E, F = ellipse_coefs.T
    phi = 1/2 * np.arctan(B/(A - C))

    c = np.cos(phi)
    s = np.sin(phi)
    c2, s2 = c**2, s**2

    A1 = A*c2 + B*c*s + C*s2
    D1 = D*c + E*s
    E1 = E*c - D*s
    C1 = A*s2 - B*c*s + C*c2
    F11 = F - ((D1**2)/(4*A1) + (E1**2)/(4*C1))

    Xc = -D1/(2*A1)
    Yc = -E1/(2*C1)
    aa = np.sqrt(-F11/A1)
    bb = np.sqrt(-F11/C1)

    # If phi needs to change then we need to recompute
    aa_le_bb = aa <= bb
    idx0 = np.logical_and(aa_le_bb, Xc < 0)
    phi[idx0] -= np.pi*np.sign(phi[idx0])

    Yc_ge_0 = Yc >= 0
    idx1 = np.logical_and(~aa_le_bb, Yc_ge_0)
    phi[idx1] += np.pi/2

    idx2 = np.logical_and(~aa_le_bb, ~Yc_ge_0)
    phi[idx2] -= np.pi/2

    idx = np.logical_or(idx0, np.logical_or(idx1, idx2))
    c[idx] = np.cos(phi[idx])
    s[idx] = np.sin(phi[idx])
    c2[idx], s2[idx] = c[idx]**2, s[idx]**2
    A1[idx] = A[idx]*c2[idx] + B[idx]*c[idx]*s[idx] + C[idx]*s2[idx]
    D1[idx] = D[idx]*c[idx] + E[idx]*s[idx]
    E1[idx] = E[idx]*c[idx] - D[idx]*s[idx]
    C1[idx] = A[idx]*s2[idx] - B[idx]*c[idx]*s[idx] + C[idx]*c2[idx]
    F11[idx] = F[idx] - ((D1[idx]**2)/(4*A1[idx]) + (E1[idx]**2)/(4*C1[idx]))
    Xc[idx] = -D1[idx]/(2*A1[idx])
    Yc[idx] = -E1[idx]/(2*C1[idx])
    aa[idx] = np.sqrt(-F11[idx]/A1[idx])
    bb[idx] = np.sqrt(-F11[idx]/C1[idx])

    # Decide sign of first term of b
    if isinstance(alpha, np.ndarray):
        bsign = np.ones(alpha.shape)
        bsign[alpha > np.arccos(np.exp(-TR/T1_guess))] = -1
    else:
        if alpha > np.arccos(np.exp(-TR/T1_guess)):
            bsign = -1
        else:
            bsign = 1

    # Compute interesting values
    Xc2 = Xc*Xc
    bb2 = bb*bb
    b = (bsign*Xc*aa + bb*np.sqrt(Xc2 - aa*aa + bb2))/(Xc2 + bb2)
    b2 = b*b
    a = bb/(b*bb + Xc*np.sqrt(1 - b2))
    ab = a*b
    M = Xc*(1 - b2)/(1 - ab)

    Mmap = np.zeros(np.prod(sh))
    Mmap[recon_idx] = M

    T2 = np.zeros(np.prod(sh))
    T2[recon_idx] = -TR/np.log(a)

    T1 = np.zeros(np.prod(sh))
    calpha = np.cos(alpha)
    T1[recon_idx] = -TR/np.log(
        ((a*(1 + calpha - ab*calpha) - b)/(a*(1 + calpha - ab) - b*calpha)))

    # compute off-resonance estimates
    Xn = I.real*c[..., None] + I.imag*s[..., None]
    Yn = I.imag*c[..., None] - I.real*s[..., None]
    Coef = (a - b)/(a*np.sqrt(1 - bb2))  # Coef=A/B
    TanT = Coef[..., None]*(Yn - Yc[..., None])/(Xn - Xc[..., None])
    T = np.arctan(TanT)  # defined on [-pi/2, pi/2]
    # unwrapping of T(k) to [-pi,pi]
    idx0 = Xn < Xc[..., None]
    idx_greater = np.logical_and(idx0, (Yn - Yc[..., None]) >= 0)
    # T[idx_greater] = np.pi - T
    np.putmask(T, idx_greater, np.pi - T[idx_greater])
    idx_lesser = np.logical_and(idx0, (Yn - Yc[..., None]) < 0)
    # T[idx_lesser] = -1*np.pi + T
    np.putmask(T, idx_lesser, -1*np.pi + T[idx_lesser])
    CosT = np.cos(T)
    c = (CosT - b[..., None])/(b[..., None]*CosT - 1)
    A = np.vstack((np.cos(pcs), np.sin(pcs))).T
    x = np.linalg.lstsq(A, c.T, rcond=None)[0]
    df = np.zeros(T1.shape)
    df[recon_idx] = -1*np.arctan2(x[1, :], x[0, :])/(2*np.pi*TR)

    if ret_all:
        # pack result matrices
        Phimap = np.zeros(np.prod(sh))
        Phimap[recon_idx] = phi
        Xcmap = np.zeros(np.prod(sh))
        Xcmap[recon_idx] = Xc
        Ycmap = np.zeros(np.prod(sh))
        Ycmap[recon_idx] = Yc
        Amap = np.zeros(np.prod(sh))
        Amap[recon_idx] = a
        Bmap = np.zeros(np.prod(sh))
        Bmap[recon_idx] = b

        return(
            np.reshape(Mmap, sh),
            np.reshape(T1, sh),
            np.reshape(T2, sh),
            np.reshape(df, sh),
            Phimap,
            Xcmap,
            Ycmap,
            Amap,
            Bmap,
        )

    return (
        np.reshape(Mmap, sh),
        np.reshape(T1, sh),
        np.reshape(T2, sh),
        np.reshape(df, sh),
    )


if __name__ == '__main__':
    pass



'''Show post-processing of data combined using robustcc.'''

from time import time

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import multivariate_normal
from skimage.filters import threshold_li
from phantominator import shepp_logan

from ssfp import bssfp, robustcc, gs_recon, planet

if __name__ == '__main__':

    # Load data
    data = np.load('/home/nicholas/Documents/research/ellipse_test/17.npy')
    data = data.transpose((0, 1, 3, 2))
    data = np.fft.fftshift(np.fft.fft2(data, axes=(0, 1)), axes=(0, 1))
    pdx2 = data.shape[0] // 4
    data = data[pdx2:-pdx2, ...]
    sx, sy, npcs, nc = data.shape[:]
    TR, alpha = 6e-3, np.deg2rad(70)

    # MR params for bSSFP sim
    pcs = np.linspace(0, 2*np.pi, npcs, endpoint=False)

    # Start timer
    t0 = time()

    # Coil combine (SOS + simple phase)
    res_rcc_simple = robustcc(data, method='simple', coil_axis=-1, pc_axis=-2)

    # Make a mask
    gs = np.array([gs_recon(res_rcc_simple[..., ii::4], pc_axis=-1) for ii in range(npcs // 4)])
    gs = np.abs(np.mean(gs, axis=0))
    thresh = threshold_li(gs)
    mask = gs > thresh

    # PLANET
    _Meff, T1map, T2map = planet(res_rcc_simple, alpha, TR, T1_guess=1, mask=mask)

    # Stop timer
    print('Recon took %g sec' % (time() - t0))

    # Take a look
    plt.subplot(1, 2, 1)
    plt.imshow(T1map, vmin=0, vmax=1)
    plt.title('Est T1')
    plt.axis('off')

    plt.subplot(1, 2, 2)
    plt.imshow(T2map, vmin=0, vmax=1)
    plt.title('Est T2')
    plt.axis('off')

    plt.show()



'''Python implementation of region growing phase correction.'''

import logging

import numpy as np
from tqdm import trange

from ssfp.utils import choose_cntr

def scatterangle(xi, yi):
    '''
	Parameters
    ----------
	xi, yi : array_like
        Vectors of data point coordinates.

	Returns
    -------
    ang : array_like
		Angle of best fit line through origin, in radians.

    Notes
    -----
	Function fits the best line through a 2D scatter of
	data, minimizing the squared closest distance from each data
	point to the line.
    '''

    sxx = np.sum(xi.flatten()**2)
    syy = np.sum(yi.flatten()**2)
    sxy = np.sum((xi*yi).flatten())

    # The equation that minimizes the sum of squared distances
    # is 0.5*tan(2*theta) = sxy / (sxx-syy).
    #
    # There is some ambiguity, since the theta that MAXIMIZES
    # the sum of squared distances is also a solution.
    #
    # However, since atan2(y,x) wraps only every 2*pi, this means
    # that theta is uniquely determined within a range of pi.
    # The solution with atan2(-y,-x) would give the theta that
    # maximizes the sum of squared distances.

    return 0.5*np.arctan2(2*sxy, sxx - syy)

def rgphcorr3d(
        im, cellsize=(4, 4, 4), use_ctr=False, slice_axis=-1,
        ret_extra=False):
    '''Region-growing phase correction for 3d complex image data.

    Parameters
    ----------
    im : array_like, 3-dimensional
        Array of complex pixel values from an SSFP acquisition.
    cellsize : list_like, optional
        Size of cubic region cells.
    use_ctr : bool, optional
        Automatically choose the center point as start.
        If use_ctr=False, use GUI to choose starting point in fat.
    slice_axis : int, optional
        Axis holding slices.  Only used if use_ctr=False.
    ret_extra : bool, optional
        Also return angles and weights.

    Returns
    -------
    pcim : array_like
        Image following phase correction.
    cellangle : array_like, optional
        Angle removed from each cell.
    cellweight : array_like, optional
        Weights for each cell (unused in current implementation).

    Notes
    -----
    We assume that the phase component of im consists
    of a slowly-spatially-varying component added to
    a random sign flip.  Beginning with a central, high-signal
    pixel, regions are added, and the slowly-varying phase
    is removed on a cell-by-cell basis.

    Implements the algorithm described in [1]_.  Based on the MATLAB
    implementation found at [2]_.

    References
    ----------
    .. [1] Hargreaves, Brian A., et al. "Fat‐suppressed steady‐state
           free precession imaging using phase detection." Magnetic
           Resonance in Medicine: An Official Journal of the
           International Society for Magnetic Resonance in Medicine
           50.1 (2003): 210-213.
    .. [2] http://mrsrl.stanford.edu/~brian/psssfp/
    '''

    assert im.ndim == 3, 'im should be a 3-dimensional array!'
    assert len(cellsize) == 3, (
        'cellsize should be a tuple with 3 entries!')
    logging.debug('cellsize is %s', cellsize)
    cellsize = np.array(cellsize)

    # ======= Find Starting Cell by Displaying Image. ===========
    szim = np.array(im.shape[:])

    # Select pixel in fat
    if use_ctr:
        # Just start in image center.
        cntr = np.round(szim/2)
    else:
        cntr = choose_cntr(np.abs(im), slice_axis=slice_axis)

    # Central cell, in cell coordinates.
    ccell = np.floor(cntr/cellsize + 1)

    # Duplicate array, for offset calcs.
    ccellbig = np.ones((27, 1))*ccell[None, :]

    # ======= Make padded array with integer # of cells =======

    ncells = np.ceil(szim/cellsize).astype(int)
    for k in range(3): # If 2x2x2, neighbour checks fail.
        if ncells[k] < 3:
            ncells[k] = 3

    # Pad the image if needed
    did_pad = False
    if not np.allclose(ncells*cellsize, szim):
        maxvox = ncells*cellsize
        logging.info(
            'Padding the image to <%d, %d, %d>',
            maxvox[0], maxvox[1], maxvox[2])

        # Do the actual padding
        px, py, pz = [x - y for x, y in zip(maxvox, im.shape)]
        adjx, adjy, adjz = np.mod([px, py, pz], 2)
        px2, py2, pz2 = int(px/2), int(py/2), int(pz/2)
        im = np.pad(
            im,
            ((px2+adjx, px2), (py2+adjy, py2), (pz2+adjz, pz2)),
            mode='constant')
        did_pad = True

    # ======= Output Variables =======

    # Allocate phase-corrected image.
    pcim = -.0001*np.ones(im.shape, dtype=im.dtype)

    # Angles ultimately removed in correction:
    cellangle = np.zeros(ncells)

    # # Weights for each cell, calculated as sum of mags of dot
    # # products with angle.
    cellweight = np.zeros(ncells)

    # ======= Define coordinate arrays =======
    #
    # These are just arrays in the cell coordinates containing
    # the x, y and z locations of each block. When cells are reordered
    # based on distance from center, these maps will be used to get
    # the cell coordinates where the calculated angle is to be placed.

    # locations of cells, in cell coords
    xmap = np.zeros(ncells, dtype=int)
    ymap = np.zeros(ncells, dtype=int)
    zmap = np.zeros(ncells, dtype=int)

    # X locations.
    # m1 = [1:ncells(1)]'*ones(1,ncells(2));
    m1 = np.arange(1, ncells[0]+1)[:, None] @ np.ones((1, ncells[1]))
    for k in range(ncells[2]):
        xmap[..., k] = m1

    # Y locations.
    # m1 = ones(ncells(1),1)*[1:ncells(2)];
    m1 = np.ones((ncells[0], 1)) @ np.arange(1, ncells[1]+1)[None, :]
    for k in range(ncells[2]):
        ymap[..., k] = m1

   # Z locations.
    # m1 = ones(ncells(2),1)*[1:ncells(3)];
    m1 = np.ones((ncells[1], 1)) @ np.arange(1, ncells[2]+1)[None, :]
    for k in range(ncells[0]):
        zmap[k, ...] = m1

    logging.debug(
        'xmap max/min - %d , %d',
        np.min(xmap.flatten()),
        np.max(xmap.flatten()))
    logging.debug(
        'ymap max/min - %d , %d',
        np.min(ymap.flatten()),
        np.max(ymap.flatten()))
    logging.debug(
        'zmap max/min - %d , %d',
        np.min(zmap.flatten()),
        np.max(zmap.flatten()))

    # ======= Figure out Distances of cells from center cell. =======
    #
    #	Calculate distances of cells from center cell, and grow
    #	region outward.  This guarantees that whenever the phase
    #	in a cell is calculated, then the phases of all closer
    #	cells have already been calculated.
    #

    # Make list of coordinates. Use Fortran ordering to match MATLAB
    # concatenation result
    cells = np.concatenate((
        xmap.flatten('F')[:, None],
        ymap.flatten('F')[:, None],
        zmap.flatten('F')[:, None]), axis=1)
    ccelle = np.ones((np.prod(ncells), 1)) @ ccell[None, :]

    # Delta, in cell coords, from center to cell.
    celldelta = cells - ccelle

    # Euclidean distance.
    celld = np.sqrt(np.sum(celldelta*celldelta, axis=-1))

    # Order distances from center, use mergesort to match MATLAB's
    # stable quicksort, see https://stackoverflow.com/questions/
    # 39484073/matlab-sort-vs-numpy-argsort-how-to-match-results
    cellorder = np.argsort(celld, kind='mergesort')

    # ======= Do phase fit and correction for cells in order. ========

    # Allocate "working cell"
    celldata = np.zeros(cellsize.shape, dtype=im.dtype)

    # Arrays of indices of starting corner for cell data in im.
    # cellstartX = ([1:ncells(1)]-1)*cellsize(1)
    # cellstartY = ([1:ncells(2)]-1)*cellsize(2)
    # cellstartZ = ([1:ncells(3)]-1)*cellsize(3)
    cellstartX = np.arange(ncells[0])*cellsize[0]
    cellstartY = np.arange(ncells[1])*cellsize[1]
    cellstartZ = np.arange(ncells[2])*cellsize[2]

    # Offsets for cells. (+1 to match MATLAB indexing)
    celloffsetX = np.arange(cellsize[0]) + 1
    celloffsetY = np.arange(cellsize[1]) + 1
    celloffsetZ = np.arange(cellsize[2]) + 1

    # ====== Main Loop Initialization ========

    # Offsets to neighbours.
    nbr_offsts = np.array(
        [
            [1, 1, 1],
            [2, 1, 1],
            [3, 1, 1],
            [1, 2, 1],
            [2, 2, 1],
            [3, 2, 1],
            [1, 3, 1],
            [2, 3, 1],
            [3, 3, 1],
            [1, 1, 2],
            [2, 1, 2],
            [3, 1, 2],
            [1, 2, 2],
            [3, 2, 2],
            [1, 3, 2],
            [2, 3, 2],
            [3, 3, 2],
            [1, 1, 3],
            [2, 1, 3],
            [3, 1, 3],
            [1, 2, 3],
            [2, 2, 3],
            [3, 2, 3],
            [1, 3, 3],
            [2, 3, 3],
            [3, 3, 3],
        ])
    nbr_offsts = 2*np.ones((26, 3)) - nbr_offsts
    nbr_dists = np.sqrt(np.sum(nbr_offsts*nbr_offsts, axis=-1))

    # ====== Main Loop =========

    # Do this for all cells.
    desc = 'Fitting cells'
    for k in trange(int(np.prod(ncells)), desc=desc, leave=False):

        cellnum = cellorder[k]
        cellnum_idx = np.unravel_index(cellnum, ncells, order='F')

    	# ===== Extract Data Points from im.
        X, Y, Z = np.meshgrid(
            cellstartX[xmap[cellnum_idx]-1] + celloffsetX - 1,
            cellstartY[ymap[cellnum_idx]-1] + celloffsetY - 1,
            cellstartZ[zmap[cellnum_idx]-1] + celloffsetZ - 1)
        celldata = im[X, Y, Z]

    	#disp('CELL RANGE: ');
    	#disp(cellstartX(xmap(cellnum))+celloffsetX);
    	#disp(cellstartY(ymap(cellnum))+celloffsetY);
    	#disp(cellstartZ(zmap(cellnum))+celloffsetZ);

    	# Calculate angle of best fit line through points, and origin.
    	# Angle is in the range [-pi/2, pi/2]
        an = scatterangle(celldata.real, celldata.imag)
        anv = np.exp(1j*an)

        # Transposition magic to match MATLAB flattening of 3-d array
        tmp = celldata.transpose((1, 0, 2)).flatten('F')
        tmp = np.concatenate((
            tmp.real[:, None],
            tmp.imag[:, None]), axis=1)
        magv = tmp @ np.array([anv.real, anv.imag])
        cellweight[cellnum_idx] = np.sum(np.abs(magv))

    	# ===== Compare with weighted average of neighbors.

        # Dot products.
        ninds = nbr_offsts @ celldelta[cellnum, :]

        # Nec. for closer.
        f = ninds < 0
        cands = nbr_offsts[f, :]
        canddists = nbr_dists[f]
        ndelts = cands + (
            np.ones((np.sum(f), 1)) @ celldelta[cellnum, :][None, :])
        ndists = np.sqrt(np.sum(ndelts*ndelts, axis=-1))

    	# f = find(ndists < celld(cellnum));
        f = ndists < celld[cellnum]
        ndelts = ndelts[f, :] # Keep only closer cell-deltas.
        nseps = canddists[f] # Sep from cell to neighbour.

        # cell coords of nbrs.
        ncoords = (ndelts + ccellbig[:np.sum(f), :]).astype(int)

		# Figure out the "average" angle of these by summing
		# a vector for each neighbour, whose length is reduced
		# by the distance separating the neighbour from the cell
		# of interest.
        vtot = 0
        if np.sum(f) > 0:
            logging.debug(
                'Cell at <%2d,%2d,%2d> ',
                cells[cellnum, 0],
                cells[cellnum, 1],
                cells[cellnum, 2])

            for p in range(np.sum(f)):
                logging.debug(
                    'Nearer Neighbour at <%2d,%2d,%2d> (sep = %f) ',
                    ncoords[p, 0],
                    ncoords[p, 1],
                    ncoords[p, 2],
                    nseps[p])

                if (all(ncoords[p, :] > 0) and
                        all(ncoords[p, :] <= ncells)):
                    vtot += 1/nseps[p]*np.exp(1j*cellangle[
                        ncoords[p, 0] - 1,
                        ncoords[p, 1] - 1,
                        ncoords[p, 2] - 1])

            # "Average" neighbour angle.
            nang = np.arctan2(vtot.imag, vtot.real)

        else:
            # At Center point -- just take angle of average of scatter
            avgvec = np.mean(celldata.flatten())
            nang = np.arctan2(avgvec.imag, avgvec.real)

    	# ===== Compare with phase of neighbors, and make sure
    	#	ultimate phase differs by less than pi/2.  The hope
    	#	here is to be insensitive to phase fluctuations of
    	#	+/- pi in the image, but only detect the slowly-varying
    	#	phase.

        # Angle difference.
        diff = an - nang

        # Modulus that to +/-(pi/2)
        rd = np.mod(diff + np.pi/2, np.pi) - np.pi/2

        # Keep |difference| < pi/2.
        corran = nang + rd

        # Keep ultimate angle in [-pi,pi]
        corran = np.mod(corran + np.pi, 2*np.pi) - np.pi

        logging.debug(
            'Cell <%3d,%3d,%3d>, fit=%3d deg.  neighbor=%3d deg.  '
            'final=%3d deg.',
            xmap[cellnum_idx],
            ymap[cellnum_idx],
            zmap[cellnum_idx],
            np.round(180/np.pi*an),
            np.round(180/np.pi*nang),
            np.round(180/np.pi*corran))

    	# ===== Store and Correct phase
        cellangle[cellnum_idx] = corran

        # Correction phase factor.
        pf = np.exp(-1j*corran)
        pcim[X, Y, Z] = pf*celldata

    # End of main loop.

    # Remove padding if we added it:
    if did_pad:
        nx, ny, nz = pcim.shape[:]
        pcim = pcim[px2+adjx:nx-px2, py2+adjy:ny-py2, pz2+adjz:nz-pz2]
        cellangle = cellangle[
            px2+adjx:nx-px2, py2+adjy:ny-py2, pz2+adjz:nz-pz2]

    if ret_extra:
        return(pcim, cellangle, cellweight)
    return pcim



'''Compare bSSFP and GRE constrasts.'''

import numpy as np
import matplotlib.pyplot as plt
from phantominator import shepp_logan

from ssfp import bssfp, spoiled_gre

if __name__ == '__main__':

    N = 128
    M0, T1, T2 = shepp_logan((N, N, 1), MR=True, zlims=(-.25, -.25))
    M0, T1, T2 = M0[..., 0], T1[..., 0], T2[..., 0]

    TR = 6e-3
    TE = TR/2
    alpha = np.deg2rad(70)
    bssfp_res = np.abs(bssfp(T1, T2, TR, alpha=alpha, M0=M0))
    gre_res = spoiled_gre(T1, T2, TR, TE, alpha=alpha, M0=M0)

    nx, ny = 1, 3
    plt.subplot(nx, ny, 1)
    plt.imshow(bssfp_res)
    plt.title('bSSFP')

    plt.subplot(nx, ny, 2)
    plt.imshow(gre_res)
    plt.title('GRE')

    plt.subplot(nx, ny, 3)
    idx = np.logical_and(bssfp_res > 0, gre_res > 0)
    plt.plot(bssfp_res[idx]/gre_res[idx])
    plt.ylim([0, 1])
    plt.title('bSSFP/GRE')

    plt.show()



'''Show post-processing of data combined using robustcc.'''

from time import time

import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import multivariate_normal
from phantominator import shepp_logan

from ssfp import bssfp, robustcc, planet

def _gaussian_csm(sx, sy, ncoil, sigma=1):
    '''Simple coil model for demo.'''
    X, Y = np.meshgrid(
        np.linspace(-1, 1, sx), np.linspace(-1, 1, sy))
    pos = np.stack((X[..., None], Y[..., None]), axis=-1)
    csm = np.empty((sx, sy, ncoil))
    cov = [[sigma, 0], [0, sigma]]
    for ii in range(ncoil):
        mu = [np.cos(ii/ncoil*np.pi*2), np.sin(ii/ncoil*2*np.pi)]
        csm[..., ii] = multivariate_normal(mu, cov).pdf(pos)
    return csm + 1j*csm

if __name__ == '__main__':

    # Sim params
    N, ncoils, npcs = 256, 64, 8
    TR, alpha = 4e-3, np.deg2rad(120)

    # Create coil sensitivities
    mps = _gaussian_csm(N, N, ncoils)

    # MR params for bSSFP sim
    pcs = np.linspace(0, 2*np.pi, npcs, endpoint=False)
    M0, T1, T2 = shepp_logan((N, N, 1), MR=True, zlims=(-.25, .25))
    M0, T1, T2 = np.squeeze(M0), np.squeeze(T1), np.squeeze(T2)

    # Linear off resonance -- exagerate off-resonance effects
    df, _ = np.meshgrid(
        np.linspace(-1/TR, 1/TR, N),
        np.linspace(-1/TR, 1/TR, N))

    # Simulate the bSSFP acquisition
    data = np.empty((N, N, npcs, ncoils), dtype=np.complex64)
    for ii in range(ncoils):
        res = bssfp(
            T1, T2, TR=TR, alpha=alpha, field_map=df, phase_cyc=pcs,
            M0=M0, delta_cs=0, phi_rf=np.angle(mps[None, ..., ii]),
            phi_edd=0, phi_drift=0, target_pc_axis=-1)
        data[..., ii] = np.abs(mps[..., None, ii])*res

    # Add noise
    sigma = 1e-7
    data += (np.random.normal(0, sigma, size=data.shape) +
             1j*np.random.normal(0, sigma, size=data.shape))

    # Start timer
    t0 = time()

    # Coil combine (SOS + simple phase)
    res_rcc_simple = robustcc(data, method='simple', coil_axis=-1, pc_axis=-2)

    # PLANET
    mask = np.abs(M0) > 1e-8
    _Meff, T1map, T2map = planet(res_rcc_simple, alpha, TR, T1_guess=1, mask=mask, pc_axis=-1)

    # Stop timer
    print('Recon took %g sec' % (time() - t0))

    # Take a look
    plt.subplot(2, 3, 1)
    plt.imshow(T1)
    plt.title('True T1')
    plt.axis('off')

    plt.subplot(2, 3, 2)
    plt.imshow(T1map)
    plt.title('Est T1')
    plt.axis('off')

    plt.subplot(2, 3, 3)
    plt.imshow(100*(T1map - T1)/(T1 + 1e-10))
    plt.title('Percent diff')
    plt.colorbar()
    plt.axis('off')

    plt.subplot(2, 3, 4)
    plt.imshow(T2)
    plt.title('True T2')
    plt.axis('off')

    plt.subplot(2, 3, 5)
    plt.imshow(T2map)
    plt.title('Est T2')
    plt.axis('off')

    plt.subplot(2, 3, 6)
    plt.imshow(100*(T2map - T2)/(T2 + 1e-10))
    plt.title('Percent diff')
    plt.colorbar()
    plt.axis('off')

    plt.show()



'''Basic usage of region growing phase correction.'''

import numpy as np
import matplotlib.pyplot as plt
from scipy.io import loadmat

from ssfp import rgphcorr3d
from ssfp.utils import download_file

if __name__ == '__main__':

    # Load sample data
    filepath = download_file(
        'http://mrsrl.stanford.edu/~brian/psssfp/samplepsssfp.mat',
        'samplepsssfp.mat')
    data = loadmat(filepath)['dat']

    # Do the phase correction
    pcim = rgphcorr3d(data, cellsize=(9, 6, 6), slice_axis=0)

    # Choose a slice to look at
    sl = 100
    plt.subplot(1, 2, 1)
    plt.imshow(np.real(data[sl, ...]))
    plt.axis('off')
    plt.title('Real(Image)')

    plt.subplot(1, 2, 2)
    plt.imshow(np.real(pcim[sl, ...]))
    plt.axis('off')
    plt.title('Real(Phase Corrected Image)')
    plt.show()

    # Look at water and fat images
    water_mask = pcim[sl, ...].real >= 0
    fat_mask = pcim[sl, ...].real < 0
    plt.subplot(1, 2, 1)
    plt.imshow(water_mask*np.abs(data[sl, ...]))
    plt.axis('off')
    plt.title('Water Image')

    plt.subplot(1, 2, 2)
    plt.imshow(fat_mask*np.abs(data[sl, ...]))
    plt.axis('off')
    plt.title('Fat Image')
    plt.show()



'''Python implementation of the PLANET algorithm.'''

import numpy as np

def planet(
        I, alpha, TR, T1_guess, pcs=None, fit_ellipse=None,
        compute_df=False, disp=False):
    '''Simultaneous T1, T2 mapping using phase‐cycled bSSFP.

    Parameters
    ----------
    I : array_like
        Complex voxels from phase-cycled bSSFP images.
    alpha : float
        Flip angle (in rad).
    TR : float
        Repetition time (in sec).
    T1_guess : float
        Estimate of expected T1 value (in sec).
    pcs : array_like, optional
        Phase-cycles that generate phase-cycle images of I (required
        if computing df) (in rad).
    fit_ellipse : callable, optional
        Function used to fit data points to ellipse.
    compute_df : bool, optional
        Whether or not estimate local off-resonance, df.
    disp : bool, optional
        Show debug plots.

    Returns
    -------
    Meff : array_like
        Effective magnetization amplitude (arbitrary units).
    T1 : array_like
        Estimate of T1 values (in sec).
    T2 : array_like
        Estimate of T2 values (in sec).
    df : array_like, optional
        Estimate of off-resonance values (in Hz).

    Raises
    ------
    AssertionError
        If fit_ellipse returns something that is not an ellipse
    AssertionError
        If the rotation fails and xc < 0 or yc =/= 0.
    AssertionError
        If a, b, or Meff are outside of interval (0, 1).
    ValueError
        If ellipse callapses to a line.
    ValueError
        If the sign of b cannot be determined.

    Notes
    -----
    Requires at least 6 phase cycles to fit the ellipse.  The ellipse
    fitting method they use (and which is implemented here) may not
    be the best method, but it is quick.  Could add more options for
    fitting in the future.

    fit_ellipse(x, y) should take two arguments and return a vector
    containing the coefficients of the implicit ellipse equation.  If
    fit_ellipse=None then the fit_ellipse_halir() function will be
    used.

    pcs should be a list of phase-cycles in radians.  If pcs=None, it
    will be determined as I.size equally spaced phasce-cycles on the
    interval [0, 2pi).

    Implements algorithm described in [1]_.

    References
    ----------
    .. [1] Shcherbakova, Yulia, et al. "PLANET: an ellipse fitting
           approach for simultaneous T1 and T2 mapping using
           phase‐cycled balanced steady‐state free precession."
           Magnetic resonance in medicine 79.2 (2018): 711-722.
    '''

    # Make sure we have an ellipse fitting function
    if fit_ellipse is None:
        fit_ellipse = _fit_ellipse_halir

    # Make sure we know what phase-cycles we have if we're computing
    # df
    if compute_df:
        if pcs is None:
            pcs = np.linspace(0, 2*np.pi, I.size, endpoint=False)
        else:
            # Make sure we get phase-cycles as a numpy array
            pcs = np.array(pcs)
        assert pcs.size == I.size, ('Number of phase-cycles must '
                                    'match entries of I!')

    ## Step 1. Direct linear least squares ellipse fitting to
    ## phase-cycled bSSFP data.
    C = fit_ellipse(I.real, I.imag)

    # Look at it in standard form
    C1, C2, C3, _C4, _C5, _C6 = C[:]
    assert C2**2 - 4*C1*C3 < 0, 'Not an ellipse!'

    ## Step 2. Rotation of the ellipse to initial vertical conic form.
    xr, yr, Cr, _phi = _do_planet_rotation(I)
    I0 = xr + 1j*yr
    xc, yc = _get_center(Cr)

    # Look at it to make sure we've rotated correctly
    if disp:
        import matplotlib.pyplot as plt
        Idraw = np.concatenate((I, [I[0]]))
        I0draw = np.concatenate((I0, [I0[0]]))
        plt.plot(Idraw.real, Idraw.imag, label='Sampled')
        plt.plot(I0draw.real, I0draw.imag, label='Rotated')
        plt.legend()
        plt.axis('square')
        plt.show()

    # Sanity check: make sure we got what we wanted:
    assert np.allclose(yc, 0), 'Ellipse rotation failed! yc = %g' % yc
    assert xc > 0, ('xc needs to be in the right half-plane! xc = %g'
                    '' % xc)
    # C1r, C2r, C3r, C4r, C5r, C6r = Cr[:]


    ## Step 3. Analytical solution for parameters Meff, T1, T2.
    # Get the semi axes, AA and BB
    A, B = _get_semiaxes(Cr)
    # Ellipse must be vertical -- so make the axes look like it
    if A > B:
        A, B = B, A
    A2 = A**2
    B2 = B**2

    # Decide sign of first term of b
    E1 = np.exp(-TR/T1_guess)
    aE1 = np.arccos(E1)
    if alpha > aE1:
        val = -1
    elif alpha < aE1:
        val = 1
    elif alpha == aE1:
        raise ValueError('Ellipse is a line! x = Meff')
    else:
        raise ValueError(
            'Houston, we should never have raised this error...')

    # See Appendix
    # xc = np.abs(xc) # THIS IS NOT IN THE APPENDIX but by def in
    # eq [9]
    xc2 = xc**2
    xcA = xc*A
    b = (val*xcA + np.sqrt(xcA**2 - (xc2 + B2)*(A2 - B2)))/(xc2 + B2)
    b2 = b**2
    a = B/(xc*np.sqrt(1 - b2) + b*B)
    ab = a*b
    Meff = xc*(1 - b2)/(1 - ab)

    # Sanity checks:
    assert 0 < b < 1, '0 < b < 1 has been violated! b = %g' % b
    if a > 1:
        a = .99
    #assert 0 < a < 1, '0 < a < 1 has been violated! a = %g' % a
    assert 0 < Meff < 1, (
        '0 < Meff < 1 has been violated! Meff = %g' % Meff)

    # Now we can find the things we were really after
    ca = np.cos(alpha)
    T1 = -1*TR/(
        np.log((a*(1 + ca - ab*ca) - b)/(a*(1 + ca - ab) - b*ca)))
    T2 = -1*TR/np.log(a)

    ## Step 4. Estimation of the local off-resonance df.
    if compute_df:
        # The beta way:
        # costheta = np.zeros(dphis.size)
        # for nn in range(dphis.size):
        #     x, y = I0[nn].real, I0[nn].imag
        #     tanbeta = y/(x - xc)
        #     t = np.arctan(A/B*tanbeta)
        #     costheta[nn] = (np.cos(t) - b)/(b*np.cos(t) - 1)

        # The atan2 way:
        costheta = np.zeros(pcs.size)
        for nn in range(pcs.size):
            x, y = I0[nn].real, I0[nn].imag
            t = np.arctan2(y, x - xc)

            if a > b:
                costheta[nn] = (np.cos(t) - b)/(b*np.cos(t) - 1)
            else:
                # Sherbakova doesn't talk about this case in the
                # paper!
                costheta[nn] = (np.cos(t) + b)/(b*np.cos(t) + 1)

        # Get least squares estimate for K1, K2
        X = np.array([np.cos(pcs), np.sin(pcs)]).T
        K = np.linalg.multi_dot((np.linalg.pinv(
            X.T.dot(X)), X.T, costheta))
        K1, K2 = K[:]

        # And finally...
        theta0 = np.arctan2(K2, K1)
        df = -1*theta0/(2*np.pi*TR) # spurious negative sign, bug!
        return(Meff, T1, T2, df)

    # else...
    return(Meff, T1, T2)

def _get_semiaxes(c):
    '''Solve for semi-axes of the cartesian form of ellipse equation.

    Parameters
    ----------
    c : array_like
        Coefficients of general quadratic polynomial function for
        conic functions.

    Returns
    -------
    float
        Semi-major axis
    float
        Semi-minor axis

    Notes
    -----
    https://en.wikipedia.org/wiki/Ellipse
    '''
    A, B, C, D, E, F = c[:]
    B2 = B**2
    den = B2 - 4*A*C
    num = 2*(A*E**2 + C*D**2 - B*D*E + den*F)
    num *= (A + C + np.array([1, -1])*np.sqrt((A - C)**2 + B2))
    AB = -1*np.sqrt(num)/den

    # # Return semi-major axis first
    # if AB[0] > AB[1]:
        # print(AB)
        # return(AB[1], AB[0])
    return(AB[0], AB[1])

def _get_center(c):
    '''Compute center of ellipse from implicit function coefficients.

    Parameters
    ----------
    c : array_like
        Coefficients of general quadratic polynomial function for
        conic funs.

    Returns
    -------
    xc : float
        x coordinate of center.
    yc : float
        y coordinate of center.
    '''
    A, B, C, D, E, _F = c[:]
    den = B**2 - 4*A*C
    xc = (2*C*D - B*E)/den
    yc = (2*A*E - B*D)/den
    return(xc, yc)

def _rotate_points(x, y, phi, p=(0, 0)):
    '''Rotate points x, y through angle phi w.r.t. point p.

    Parameters
    ----------
    x : array_like
        x coordinates of points to be rotated.
    y : array_like
        y coordinates of points to be rotated.
    phi : float
        Angle in radians to rotate points.
    p : tuple, optional
        Point to rotate around.

    Returns
    -------
    xr : array_like
        x coordinates of rotated points.
    yr : array_like
        y coordinates of rotated points.
    '''
    x = x.flatten()
    y = y.flatten()
    xr = (x - p[0])*np.cos(phi) - (y - p[0])*np.sin(phi) + p[0]
    yr = (y - p[1])*np.cos(phi) + (x - p[1])*np.sin(phi) + p[1]
    return(xr, yr)

def _do_planet_rotation(I):
    '''Rotate complex pts to fit vertical ellipse centered at (xc, 0).

    Parameters
    ----------
    I : array_like
        Complex points from SSFP experiment.

    Returns
    -------
    xr : array_like
        x coordinates of rotated points.
    yr : array_like
        y coordinates of rotated points.
    cr : array_like
        Coefficients of rotated ellipse.
    phi : float
        Rotation angle in radians of effective rotation to get
        ellipse vertical and in the x > 0 half plane.
    '''

    # Represent complex number in 2d plane
    x = I.real.flatten()
    y = I.imag.flatten()

    # Fit ellipse and find initial guess at what rotation will make it
    # vertical with center at (xc, 0).  The arctan term rotates the
    # ellipse to be horizontal, then we need to decide whether to add
    # +/- 90 degrees to get it vertical.  We want xc to be positive,
    # so we must choose the rotation to get it vertical.
    c = _fit_ellipse_halir(x, y)
    phi = -.5*np.arctan2(c[1], (c[0] - c[2])) + np.pi/2
    xr, yr = _rotate_points(x, y, phi)

    # If xc is negative, then we chose the wrong rotation! Do -90 deg
    cr = _fit_ellipse_halir(xr, yr)
    if _get_center(cr)[0] < 0:
        # print('X IS NEGATIVE!')
        phi = -.5*np.arctan2(c[1], (c[0] - c[2])) - np.pi/2
        xr, yr = _rotate_points(x, y, phi)

    # Fit the rotated ellipse and bring yc to 0
    cr = _fit_ellipse_halir(xr, yr)
    yr -= _get_center(cr)[1]
    cr = _fit_ellipse_halir(xr, yr)
    # print(_get_center(cr))

    # With noisy measurements, sometimes the fit is incorrect in the
    # above steps and the ellipse ends up horizontal.  We can realize
    # this by finding the major and minor semiaxes of the ellipse.
    # The first axis returned should be the smaller if we were
    # correct, if not, do above steps again with an extra factor of
    # +/- 90 deg to get the ellipse standing up vertically.
    ax = _get_semiaxes(c)
    if ax[0] > ax[1]:
        # print('FLIPPITY FLOPPITY!')
        xr, yr = _rotate_points(x, y, phi + np.pi/2)
        cr = _fit_ellipse_halir(xr, yr)
        if _get_center(cr)[0] < 0:
            # print('X IS STILL NEGATIVE!')
            phi -= np.pi/2
            xr, yr = _rotate_points(x, y, phi)
        else:
            phi += np.pi/2

        cr = _fit_ellipse_halir(xr, yr)
        yr -= _get_center(cr)[1]
        cr = _fit_ellipse_halir(xr, yr)
        # print(_get_center(cr))

    return(xr, yr, cr, phi)

def _fit_ellipse_halir(x, y):
    '''Improved ellipse fitting algorithm by Halir and Flusser.

    Parameters
    ----------
    x : array_like
        y coordinates assumed to be on ellipse.
    y : array_like
        y coordinates assumed to be on ellipse.

    Returns
    -------
    array_like
        Ellipse coefficients.

    Notes
    -----
    Note that there should be at least 6 pairs of (x,y).

    From the paper's conclusion:

        "Due to its systematic bias, the proposed fitting algorithm
        cannot be used directly in applications where excellent
        accuracy of the fitting is required. But even in that
        applications our method can be useful as a fast and robust
        estimator of a good initial solution of the fitting
        problem..."

    See figure 2 from [2]_.
    '''

    # We should just have a bunch of points, so we can shape it into
    # a column vector since shape doesn't matter
    x = x.flatten()
    y = y.flatten()

    # Make sure we have at least 6 points (6 unknowns...)
    if x.size < 6 and y.size < 6:
        print((
            'WARNING: We need at least 6 sample points for a good '
            'fit!'))

    # Here's the heavy lifting
    D1 = np.stack((x**2, x*y, y**2)).T # quadratic pt of design matrix
    D2 = np.stack((x, y, np.ones(x.size))).T # lin part design matrix
    S1 = np.dot(D1.T, D1) # quadratic part of the scatter matrix
    S2 = np.dot(D1.T, D2) # combined part of the scatter matrix
    S3 = np.dot(D2.T, D2) # linear part of the scatter matrix
    T = -1*np.linalg.inv(S3).dot(S2.T) # for getting a2 from a1
    M = S1 + S2.dot(T) # reduced scatter matrix
    M = np.array([M[2, :]/2, -1*M[1, :], M[0, :]/2]) #premult by C1^-1
    _eval, evec = np.linalg.eig(M) # solve eigensystem
    cond = 4*evec[0, :]*evec[2, :] - evec[1, :]**2 # evaluate a’Ca
    a1 = evec[:, cond > 0] # eigenvector for min. pos. eigenvalue
    a = np.vstack([a1, T.dot(a1)]).squeeze() # ellipse coefficients
    return a

if __name__ == '__main__':
    pass



'''Utility functions.'''

import pathlib
import urllib.request
from math import ceil
import logging

import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm

def ernst(TR, T1):
    '''Computes the Ernst angle.

    Parameters
    ----------
    TR : float
        repetition time.
    T1 : array_like
        longitudinal exponential decay time constant.

    Returns
    -------
    alpha : array_like
        Ernst angle in rad.

    Notes
    -----
    Implements equation [14.9] from [1]_.

    References
    ----------
    .. [1] Notes from Bernstein, M. A., King, K. F., & Zhou, X. J.
           (2004). Handbook of MRI pulse sequences. Elsevier.
    '''

    # Don't divide by zero!
    alpha = np.zeros(T1.shape)
    idx = np.nonzero(T1)
    alpha[idx] = np.arccos(-TR/T1[idx])
    return alpha

def download_file(address, filename, force=False):
    '''Download a file into data folder.

    Parameters
    ----------
    address : str
        URL to get the file.
    filename : str
        Filename to save the file at URL to.
    force : bool, optional
        Download again, even if the file already exists locally.

    Returns
    -------
    path : str
        Local path to downloaded file.
    '''

    # Make sure destination directory exists
    dest = pathlib.Path('data/')
    dest.mkdir(parents=True, exist_ok=True)

    # Don't download if we already have it
    if not force and (dest / filename).exists():
        logging.info('File is already downloaded!')
        return str(dest / filename)

    # Else, get file from interwebs
    pbar = None
    def _progress(_num_blocks, block_size, file_size):
        nonlocal pbar
        if pbar is None:
            pbar = tqdm(
                desc='Downloading file',
                total=ceil(file_size/block_size), leave=False)
        pbar.update(1)

    _local_filename, _ = urllib.request.urlretrieve(
        address, str(dest / filename), _progress)
    pbar.close()
    return str(dest / filename)

class IndexTracker(object):
    '''Use scroll wheel event to cycle through slices.'''
    def __init__(self, ax, X):
        self.ax = ax
        ax.set_title('use scroll wheel to navigate slices')
        ax.set_xlabel(
            'left click to select point, right click to remove')

        self.X = X
        _rows, _cols, self.slices = X.shape
        self.ind = self.slices//2

        self.im = ax.imshow(self.X[:, :, self.ind])
        self.update()

    def onscroll(self, event):
        '''Trigger scrolling event.'''
        # print("%s %s" % (event.button, event.step))
        if event.button == 'up':
            self.ind = (self.ind + 1) % self.slices
        else:
            self.ind = (self.ind - 1) % self.slices
        self.update()

    def update(self):
        '''Load new slice.'''
        self.im.set_data(self.X[:, :, self.ind])
        self.ax.set_ylabel('slice %s' % self.ind)
        self.im.axes.figure.canvas.draw()

def choose_cntr(im, slice_axis=-1):
    '''Graphically choose point '''

    fig, ax = plt.subplots(1, 1)
    tracker = IndexTracker(ax, np.moveaxis(im, slice_axis, -1))
    fig.canvas.mpl_connect('scroll_event', tracker.onscroll)

    # Get two clicks, first is the actual point, second
    # is just to close the input window
    cntr = fig.ginput(n=2, show_clicks=True)[0]

    # Get the current slice index and create cntr coord
    zidx = tracker.ind
    cntr = (zidx, int(cntr[0]), int(cntr[1]))
    plt.close(fig)

    logging.info('Choosing center point: %s', str(cntr))
    return cntr



'''Robust coil combination for bSSFP.'''

import numpy as np
from tqdm import tqdm

def robustcc(
        data, method='simple', mask=None, coil_axis=-1, pc_axis=-2):
    '''Robust elliptcal-model-preserving coil combination for bSSFP.

    Parameters
    ----------
    data : array_like
        Complex bSSFP image space data to be coil combined.
    method : str, {'simple', 'full'}, optional
        The method of phase estimation.  method='simple' is very fast
        and does a good job.
    mask : array_like or None, optional
        Only used for method='full'.  Gives mask of which pixels to
        do fit on.  Does not have coil, phase-cycle dimensions.
    coil_axis : int, optional
        Dimension holding the coil data.
    pc_axis : int, optional
        Dimension holding the phase-cycle data.

    Returns
    -------
    res : array_like
        Complex coil-combined data that preserves elliptical
        relationships between phase-cycle pixels.

    Notes
    -----
    Implements the method described in [1]_.  This coil combination
    method preserves elliptical relationships between phase-cycle
    pixels for more efficient computation, e.g., gs_recon which
    reconstructs one coil at a time ([2]_).

    References
    ----------
    .. [1] N. McKibben, G. Tarbox, E. DiBella, and N. K. Bangerter,
           "Robust Coil Combination for bSSFP Elliptical Signal
           Model," Proceedings of the 28th Annual Meeting of the
           ISMRM; Sydney, NSW, Australia, 2020.
    .. [2] Xiang, Qing‐San, and Michael N. Hoff. "Banding artifact
           removal for bSSFP imaging with an elliptical signal model."
           Magnetic resonance in medicine 71.3 (2014): 927-933.
    '''

    # Put coil and phase-cycle axes where we expect them
    data = np.moveaxis(data, (coil_axis, pc_axis), (-1, -2))

    # Use SOS for magnitude estimate
    mag = np.sqrt(np.sum(np.abs(data)**2, axis=-1))

    # Choose the method for phase estimation
    if method == 'simple':
        # simple coil phase estimate is fast:
        phase = _simple_phase(data)
    elif method == 'full':
        # Use the full solution:
        phase = _full_phase(data, mask)
    else:
        raise NotImplementedError()

    return np.moveaxis(
        mag*np.exp(1j*phase), (-1, -2), (coil_axis, pc_axis))

def _simple_phase(data):
    '''Simple strategy: choose phase from best coil ellipse.'''

    # Assume the best coil is the one with max median value
    idx = np.argmax(
        np.median(np.abs(data), axis=-1, keepdims=True), axis=-1)
    return np.take_along_axis(
        np.angle(data), idx[..., None], axis=-1).squeeze()

def _full_phase(data, mask=None):
    '''Do pixel-by-pixel ellipse registration.'''

    # Worst case is to do all pixels:
    if mask is None:
        mask = np.ones(data.shape[:-2], dtype=bool)

    # Do pixel-by-pixel phase estimation
    phase = np.empty(data.shape[:-1])
    for idx in tqdm(
            np.argwhere(mask),
            total=np.prod(np.sum(mask.flatten())),
            leave=False):

        # Register all coil ellipses to a single reference
        coil_ellipses = data[tuple(idx) + (slice(None), slice(None))]

        # Take reference ellipse to be the one with greatest
        # median value
        ref_idx = np.argmax(np.median(np.abs(
            coil_ellipses), axis=-1, keepdims=True), axis=-1)
        ref = np.take_along_axis(
            coil_ellipses, ref_idx[..., None], axis=-1).squeeze()

        # Do coil by coil registration
        reg_ellipses = np.empty(
            coil_ellipses.shape, dtype=coil_ellipses.dtype)
        W = np.empty(coil_ellipses.shape[-1]) # weights
        for cc in range(coil_ellipses.shape[-1]):
            T = np.linalg.lstsq(
                coil_ellipses[..., cc][:, None],
                ref, rcond=None)[0]
            W[cc] = np.abs(T)**2 # save the weights
            T = np.exp(1j*np.angle(T)) # just rotate, no scaling
            reg_ellipses[..., cc] = T*coil_ellipses[..., cc]

        # Take the weighted average to the composite ellipse
        phase[tuple(idx) + (slice(None),)] = np.average(
            np.angle(reg_ellipses), axis=-1, weights=W)

    # make sure undefined values are set to 0; we have to do this
    # since we allocated with np.empty
    phase[~mask] = 0 # pylint: disable=E1130
    return phase

if __name__ == '__main__':
    pass



'''Move functions to module level.'''

from .bssfp import bssfp
from .gs_recon import gs_recon
from .planet import planet
from .gre import spoiled_gre
from .rgphcorr import rgphcorr3d
from .robustcc import robustcc



'''Show basic usage of spoiled GRE.'''

import numpy as np
import matplotlib.pyplot as plt
from phantominator import shepp_logan

from ssfp import spoiled_gre

if __name__ == '__main__':

    N = 128
    M0, T1, T2 = shepp_logan((N, N, 1), MR=True, zlims=(-.25, -.25))
    M0, T1, T2 = M0[..., 0], T1[..., 0], T2[..., 0]

    TR, TE = 0.035, 0.01
    alpha = np.deg2rad(70)
    res = spoiled_gre(T1, T2, TR, TE, alpha=alpha, M0=M0)

    plt.imshow(res)
    plt.show()



'''Look at ellipses of different harmonics.'''

import numpy as np
import matplotlib.pyplot as plt
from ssfp import bssfp

if __name__ == '__main__':

    npcs = 16
    pcs = np.linspace(0, 2*np.pi, npcs, endpoint=True)

    T1, T2, M0 = 1.2, .5, 1
    alpha = np.deg2rad(30)
    TR = 3e-3

    df0 = 3/(2*TR)
    sig0 = bssfp(T1, T2, TR, alpha, field_map=0, phase_cyc=pcs, M0=M0)
    sig1 = bssfp(
        T1, T2, TR, alpha, field_map=df0, phase_cyc=pcs, M0=M0)

    sig0 /= np.linalg.norm(sig0)
    sig1 /= np.linalg.norm(sig1)

    plt.hist([
        np.concatenate((sig0.real, sig0.imag)),
        np.concatenate((sig1.real, sig1.imag))])
    plt.show()
    # plt.hist([sig0.real, sig1.real])
    # plt.show()
    # plt.hist([sig0.imag, sig1.imag])
    # plt.show()

    df = np.linspace(-1/TR, 1/TR, npcs)
    plt.plot(df, np.abs(sig0))
    plt.plot(df, np.roll(np.abs(sig1), 250))
    plt.show()

    plt.plot(sig0.real, sig0.imag)
    plt.plot(sig1.real, sig1.imag)
    plt.axis('square')
    plt.show()

    # Notes:
    #     - Histogram of harmonics should be the same
    #     - ellipses of harmonics appear to be only slightly different



'''Try to evaluate some line integrals.'''

import numpy as np
import matplotlib.pyplot as plt
# from skimage.feature import canny
# from scipy.integrate import quad
# from tqdm import tqdm

if __name__ == '__main__':

    fft = lambda x0: np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(
        x0)))
    ifft = lambda x0: np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(
        x0)))

    a, b = 1, 2
    N = 75
    X, Y = np.meshgrid(
        np.linspace(-3, 3, N),
        np.linspace(-3, 3, N))
    Z = X**2/a**2 + Y**2/b**2 <= 1
    # Z = canny(Z)
    FZ = fft(Z)

    # Try line integral
    npcs = 500
    t = np.linspace(0, 2*np.pi, npcs, endpoint=False)
    x = a*np.cos(t)
    y = b*np.sin(t)

    def f(t, a, b, u, v):
        return np.exp(
            -1j*(u*a*np.cos(t) + v*b*np.sin(t)))*np.sqrt(
                a**2*np.sin(t)**2 + b**2*np.cos(t)**2)
    def fr(t, a, b, u, v):
        return np.real(f(t, a, b, u, v))
    def fi(t, a, b, u, v):
        return np.imag(f(t, a, b, u, v))

    u = X[..., None]*2*np.pi
    v = Y[..., None]*2*np.pi
    t = t[None, None, :]
    lFZ = np.sum(np.exp(-1j*(u*a*np.cos(t) + v*b*np.sin(t)))*np.sqrt(a**2*np.sin(t)**2 + b**2*np.cos(t)**2), axis=-1)

    # lFZ = np.zeros((N, N), dtype=FZ.dtype)
    # for idx in tqdm(np.ndindex((N, N)), total=N**2, leave=False):
    #     ii, jj = idx[:]
    #     re = quad(fr, 0, 2*np.pi, args=(
    #         a, b, X[ii, jj]*2*np.pi, Y[ii, jj]*2*np.pi))
    #     im = quad(fi, 0, 2*np.pi, args=(
    #         a, b, X[ii, jj]*2*np.pi, Y[ii, jj]*2*np.pi))
    #     lFZ[ii, jj] = re[0] + 1j*im[0]
    #
    plt.subplot(1, 2, 1)
    plt.imshow(np.abs(FZ))
    plt.subplot(1, 2, 2)
    plt.imshow(np.abs(lFZ))
    plt.show()

    plt.subplot(1, 2, 1)
    plt.imshow(Z)
    plt.subplot(1, 2, 2)
    plt.imshow(np.abs(ifft(lFZ)))
    plt.show()
